import ntpath
import os
import sys
from os import listdir
from os.path import join, isfile

import numpy as np
from PIL import Image
from keras import metrics
from keras.layers import Dense, Dropout, Flatten, Conv2D, MaxPool2D, BatchNormalization
from keras.models import Sequential
from keras.models import load_model
from keras.utils.np_utils import to_categorical
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import confusion_matrix
from mini_geant import COLLISION_TYPE

MINI_BATCH = 128
EPOCHS = 30
from keras.models import load_model

import time


def load_files(train_folder_name):
    """
    This method loads all the image files into memory and in the format of N inputs and N labels
    :param train_folder_name: path of the folder that contains the data
    :return: Images, Labels
    """
    files = []
    class_label = []
    path = join(os.getcwd(), train_folder_name)
    for file in listdir(path):
        fPath = join(path, file)
        if isfile(fPath):
            img = Image.open(fPath)
            arr = (np.array(img) / 255.0)
            files.append(arr)
            f_name = os.path.splitext(ntpath.basename(fPath))[0]
            key_list = list(COLLISION_TYPE.keys())
            val_list = list(COLLISION_TYPE.values())
            position = val_list.index(f_name.split(".")[0].split("_", 1)[1])
            class_label.append(key_list[position])
    return files, class_label


def create_model(summary):
    """
    This methods sets up the models and prints a summary of the model
    :param summary: a flag that enables/disables a summary print
    :return: model
    """
    model = Sequential()
    model.add(Conv2D(32, kernel_size=5, input_shape=(64, 64, 1), activation='relu'))
    model.add(MaxPool2D(2, 2))
    model.add(BatchNormalization())
    model.add(Flatten())
    model.add(Dense(256, activation="relu"))
    model.add(Dropout(0.4))
    model.add(Dense(4, activation="softmax"))
    model.compile(optimizer='sgd', loss="mean_squared_error", metrics=[metrics.categorical_accuracy])
    if summary:
        model.summary()
    return model


def train(model, x_train, y_train, x_val, y_val, model_file):
    """

    :param model: The model on which to train
    :param x_train: Training data in array format
    :param y_train: Labels in array format
    :param x_val: validation data in array format
    :param y_val: validation labels in array format
    :param model_file: Directory location where the model will be saved
    :return: None, only saves the model
    """
    model.fit(x_train, np.array(y_train), batch_size=MINI_BATCH, epochs=EPOCHS, validation_data=(x_val, y_val),
              verbose=1)
    model.save(model_file)
    print("Processing complete!")
    print("total number of items trained on " + str(len(x_train)))


def test(x_test, y_test, model_file, return_flag):
    """
    This method evaluates the given model for the data provided and prints the loss and accuracy.
    :param x_test: Data for testing
    :param y_test: labels of the testing data
    :param model_file: The model that mus be loaded for testing
    :return: None
    """
    # Check if model exists before trying to load it
    if not os.path.exists(model_file) or not os.path.getsize(model_file) > 0:
        print("CAN'T FIND MODEL FILE")
        exit(1)

    model = load_model(model_file)
    start = time.time()
    evaluation = model.evaluate(x_test, y_test, batch_size=MINI_BATCH, verbose=1)
    end = time.time()
    print("total number of items tested on " + str(len(x_test)))
    preds = model.predict(x_test)
    y_pred = []
    y_pred = np.argmax(preds, axis=1)
    print("Confusion Matrix")
    print(confusion_matrix(np.argmax(y_test, axis=1), y_pred))
    if return_flag:
        return evaluation[1], end - start


def f_fold(model_file, x_train, class_labels):
    """
    This method perform stratified K-fold cross-validation on the data and then performs training followed by testing on each fold
    :param model_file: location to write the model file
    :param x_train: training data
    :param class_labels: labels for the training data
    :return: None
    """
    x_values = np.array(x_train).reshape(-1, 64, 64, 1)
    y_values = np.array(class_labels)
    five_fold = StratifiedKFold(n_splits=5)
    split = 1
    avg_accuracy = 0
    training_time = 0
    testing_time = 0

    for train1, test1 in five_fold.split(x_values, y_values):
        print("Training on split:" + str(split))
        model = None
        model = create_model(True if split == 5 else False)
        start = time.time()
        train(model, x_values[train1], to_categorical(y_values[train1], num_classes=4), x_values[test1],
              to_categorical(y_values[test1], num_classes=4), model_file)
        end = time.time()
        training_time += end - start
        print("Testing on split:" + str(split))
        temp_avg_accuracy, temp_testing_time = test(x_values[test1], to_categorical(y_values[test1]), model_file, True)
        avg_accuracy += temp_avg_accuracy
        testing_time += temp_testing_time
        split += 1
    print("Total overall Accuracy over testdata: " + str(avg_accuracy / split))
    print("Total run-time for Training: " + str(training_time))
    print("Total run-time for Testing: " + str(testing_time))


if __name__ == '__main__':
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
    # fetch arguments
    if len(sys.argv) > 3:
        mode = sys.argv[1]
        model_file = sys.argv[2]
        data_folder = sys.argv[3]
    else:
        print("Incorrect number of arguments")


    # checks if training data directory exists
    if not os.path.exists(data_folder):
        print("NO TRAINING DATA")
        exit(1)

    # train data is loaded
    x_train, class_labels = load_files(data_folder)

    if mode.lower() == "train":

        # set up training and validation data
        x_train = np.array(x_train).reshape(-1, 64, 64, 1)
        y_train = to_categorical(class_labels, num_classes=4)
        x_val, class_labels = x_train[:1000], class_labels[:1000]
        x_val = np.array(x_val).reshape(-1, 64, 64, 1)
        y_val = to_categorical(class_labels, num_classes=4)

        model = None
        model = create_model(True)

        train(model, x_train, y_train, x_val, y_val, model_file)

    elif mode.lower() == "5fold":
        f_fold(model_file, x_train, class_labels)

    elif mode.lower() == "test":
        # set up training data
        x_test = np.array(x_train).reshape(-1, 64, 64, 1)
        y_test = to_categorical(class_labels, num_classes=4)

        test(x_test, y_test, model_file, False)

    else:
        print("Incorrect mode! please check and try again.")
