import os
import sys
import random
import numpy as np
from PIL import Image

COLLISION_TYPE = {
    0: "NO_COLLISION",
    1: "INELASTIC",
    2: "ELASTIC",
    3: "INELASTIC_DECAY"
}


class Particle:
    def __init__(self, size, velocity, direction):
        self.particle = [1] * random.randint(size[0], size[1]) if len(size) > 1 else [1] * size[0]
        self.direction = direction
        self.vel = random.randint(velocity[0], velocity[1]) if len(velocity) > 1 else velocity[0]
        self.pos = 0 if direction else -1

    @property
    def len(self):
        return len(self.particle)


def inelastic_collision(iterations, particle, sensor, number_of_sensors):
    """
    This method simulates a merged particle movement
    :param iterations: Number of time steps to watch for
    :param particle: Merged particle
    :param sensor: Current state of sensor
    :param number_of_sensors: THe length of sensor array
    :return: The current state of sensor array and the current position of the particle after iterations
    """
    for i in range(iterations):
        sensor_array = [0] * number_of_sensors
        for l in range(particle.len):
            if particle.pos + l < number_of_sensors:
                sensor_array[particle.pos + l] = particle.particle[l]
        particle.pos += particle.len + particle.vel
        sensor.insert(0, sensor_array)
    return sensor, particle.pos


def setup_fused_particle(p_left, p_right):
    """
    This method will fuse given 2 particle and create a new one
    :param p_left: the particle travelling from left to right
    :param p_right: the particle travelling from right to left
    :return: the new particle
    """
    velocity = int((p_left.vel * p_left.len) + (p_right.vel * p_right.len) / (
            p_left.len + p_right.len))
    p_new = Particle([p_right.len + p_left.len], [abs(velocity)], velocity > 0)
    p_new.pos = p_left.pos
    return p_new


def update_decay_flag(decay_iteration):
    """
    This method updates and checks the iterations to be skipped in case of inelastic decay collisions
    :param decay_iteration: number of decay iterations
    :return: status of decay flag
    """
    decay_iteration -= 1
    if decay_iteration == 0:
        return decay_iteration, False
    else:
        return decay_iteration, True


def setup_folder(path):
    """
    This method will check if the output directory exists it will delete all the files otherwise it will create the directory
    """
    isExist = os.path.exists(path)
    if not isExist:
        # Create a new directory because it does not exist
        os.makedirs(path)
    else:
        for f in os.listdir(path):
            fpath = os.path.join(path, f)
            try:
                if os.path.isfile(fpath):
                    os.unlink(fpath)
            except Exception as e:
                print(e)


if __name__ == '__main__':

    # fetch arguments
    if len(sys.argv) > 8:
        num_experiments = int(sys.argv[1])
        min_start_difference = int(sys.argv[2])
        max_start_difference = int(sys.argv[3])
        min_size = int(sys.argv[4])
        max_size = int(sys.argv[5])
        min_velocity = int(sys.argv[6])
        max_velocity = int(sys.argv[7])
        output_folder = sys.argv[8]
    else:
        print("Incorrect number of arguments")

    # testing
    # num_experiments = 2000
    # min_start_difference = 1
    # max_start_difference = 5
    # min_size = 1
    # max_size = 5
    # min_velocity = 1
    # max_velocity = 5
    # output_folder = "K-FOLD DATA"

    # setting parameters
    p_len = (min_size, max_size)
    p_velocity = (min_velocity, max_velocity)
    setup_folder(output_folder + "/")

    number_of_sensors = 64
    print("Generating data....")
    for exp_num in range(num_experiments):
        sensor_array = [0] * number_of_sensors
        # create particle fired from left
        p_left = Particle(p_len, p_velocity, True)
        # create particle fired from right
        p_right = Particle(p_len, p_velocity, False)
        sensor = []
        collision = 0
        file_label = str(exp_num + 1) + "_" + COLLISION_TYPE[collision]
        start_time = random.randint(min_start_difference, max_start_difference)
        decay_flag = False

        # observer for 64 time steps
        for i in range(number_of_sensors):
            # create the sensor
            sensor_array = [0] * number_of_sensors
            # check if elastic decay collision has occurred => skip steps
            if decay_flag:
                decay_iteration, decay_flag = update_decay_flag(decay_iteration)
                continue

            # start left particle movement
            for l in range(p_left.len):
                # check if particle is still inside sensor array
                if p_left.pos + l < number_of_sensors:
                    sensor_array[p_left.pos + l] = p_left.particle[l]
            p_left.pos += p_left.len + p_left.vel

            # check if right particle is allowed to move then move it
            if i > start_time:
                for r in range(-1, -p_right.len - 1, -1):
                    # check if particle is still inside sensor array
                    if p_right.pos + r > -64 and p_right.pos < 1:
                        # check for collision => if at current position sensor is already 1 => collision occurred
                        # print(str(p_right.pos + r))
                        if sensor_array[p_right.pos + r] == 1 and collision == 0:
                            # randomly selects a collision type
                            collision = random.randint(1, 3)
                            if collision == 1:
                                file_label = str(exp_num + 1) + "_" + COLLISION_TYPE[collision]
                                p_new = setup_fused_particle(p_left, p_right)
                                velocity = p_new.vel
                                sensor, p_new.pos = inelastic_collision(number_of_sensors - i, p_new, sensor,
                                                                        number_of_sensors)
                                break

                            elif collision == 2:
                                file_label = str(exp_num + 1) + "_" + COLLISION_TYPE[collision]
                                p_left.vel = int(round((p_right.len * p_right.vel) / p_left.len))
                                p_right.vel = int(round((p_left.len * p_left.vel) / p_right.len))

                            elif collision == 3:
                                file_label = str(exp_num + 1) + "_" + COLLISION_TYPE[collision]
                                p_new = setup_fused_particle(p_left, p_right)
                                velocity = p_new.vel
                                decay_iteration = random.randint(min_start_difference, max_start_difference)
                                sensor, p_new.pos = inelastic_collision(decay_iteration, p_new, sensor,
                                                                        number_of_sensors)
                                decay_flag = True
                                p_left.pos = p_new.pos
                                p_right.pos = p_new.pos - 64
                                p_left.vel = int((p_right.len * p_right.vel) / p_left.len)
                                p_right.vel = int((p_left.len * p_left.vel) / p_right.len)
                                break
                        sensor_array[p_right.pos + r] = p_right.particle[r]
                if collision == 1:
                    break
                p_right.pos -= p_right.len + p_right.vel
            sensor.insert(0, sensor_array)
        img = Image.fromarray(np.uint8(np.array(sensor) * 255), "L")
        img.save(output_folder + "/" + file_label + ".png", "PNG")
    print("Successfully generated data....")
